@@ -151,7 +151,7 @@ class BatchNorm
   //! Locally-stored epsilon value.
   double eps;
 
-  //! Whether we are in loading or saving mode.
+  //! Variable to keep track of whether we are in loading or saving mode.
   bool loading;
 
   //! Locally-stored scale parameter.
@@ -195,6 +195,9 @@ class BatchNorm
 
   //! Locally-stored normalized input.
   OutputDataType normalized;
+
+  //! Locally-stored zero mean input.
+  OutputDataType inputMean;
 }; // class BatchNorm
 
 } // namespace ann
